@@ -1,10 +1,25 @@
 package Chemistry::OpenSMILES::Stereo;
 
+# ABSTRACT: Stereochemistry handling routines
+our $VERSION = '0.11.1'; # VERSION
+
 use strict;
 use warnings;
 
-# ABSTRACT: Stereochemistry handling routines
-our $VERSION = '0.9.0'; # VERSION
+use Chemistry::OpenSMILES qw(
+    is_chiral
+    is_cis_trans_bond
+    is_double_bond
+    is_ring_bond
+    is_single_bond
+    toggle_cistrans
+);
+use Chemistry::OpenSMILES::Stereo::Tables qw( @TB );
+use Chemistry::OpenSMILES::Writer qw( write_SMILES );
+use Graph::Traversal::BFS;
+use Graph::Undirected;
+use List::Util qw( all any first max min sum sum0 uniq );
+use Set::Object qw( set );
 
 require Exporter;
 our @ISA = qw( Exporter );
@@ -15,31 +30,36 @@ our @EXPORT_OK = qw(
     mark_cis_trans
 );
 
-use Chemistry::OpenSMILES qw(
-    is_cis_trans_bond
-    is_double_bond
-    is_ring_bond
-    is_single_bond
-    toggle_cistrans
-);
-use Chemistry::OpenSMILES::Writer qw( write_SMILES );
-use Graph::Traversal::BFS;
-use Graph::Undirected;
-use List::Util qw( all any max min sum sum0 uniq );
-
 sub mark_all_double_bonds
 {
     my( $graph, $setting_sub, $order_sub, $color_sub ) = @_;
 
+    my @double_bonds = grep { is_double_bond( $graph, @$_ ) } $graph->edges;
+    if( ref $setting_sub eq 'ARRAY' ) {
+        # List of double bonds with their setting are given
+        @double_bonds = map { [ @{$_}[1..2] ] } @$setting_sub;
+        my %cis = map { ( join( '', sort @{$_}[1..2] ) => { atoms => set( $_->[0], $_->[3] ), setting => $_->[4] } ) }
+                      @$setting_sub;
+        $setting_sub = sub {
+            my $key = join '', sort @_[1..2];
+            return undef unless exists $cis{$key};
+
+            my $setting = $cis{$key}->{setting};
+            return $setting unless ($cis{$key}->{atoms} * set( $_[0], $_[3] ))->size == 1;
+            return $setting eq 'cis' ? 'trans' : 'cis';
+        };
+    }
+
     # By default, whenever there is a choice between atoms, the one with
     # lowest position in the input SMILES is chosen:
-    $order_sub = sub { return $_[0]->{number} } unless $order_sub;
+    $order_sub = sub { $_[0]->{number} } unless $order_sub;
 
     # Select non-ring double bonds
-    my @double_bonds = grep { is_double_bond( $graph, @$_ ) &&
-                              !is_ring_bond( $graph, @$_ ) &&
-                              !is_unimportant_double_bond( $graph, @$_, $color_sub ) }
-                            $graph->edges;
+    @double_bonds = grep { !is_ring_bond( $graph, @$_ ) &&
+                           !is_unimportant_double_bond( $graph, @$_, $color_sub ) }
+                         @double_bonds;
+
+    return unless @double_bonds;
 
     # Construct a double bond incidence graph. Vertices are double bonds
     # and edges are between those double bonds that separated by a single
@@ -121,7 +141,7 @@ sub mark_cis_trans
 
     # By default, whenever there is a choice between atoms, the one with
     # lowest position in the input SMILES is chosen:
-    $order_sub = sub { return $_[0]->{number} } unless $order_sub;
+    $order_sub = sub { $_[0]->{number} } unless $order_sub;
 
     my @neighbours2 = $graph->neighbours( $atom2 );
     my @neighbours3 = $graph->neighbours( $atom3 );
@@ -229,50 +249,103 @@ sub mark_cis_trans
     }
 }
 
-# Store the tetrahedral chirality character as additional pseudo vertices
-# and edges.
+# Store chirality character as additional pseudo vertices and edges.
 sub chirality_to_pseudograph
 {
     my( $moiety ) = @_;
 
     for my $atom ($moiety->vertices) {
-        next unless Chemistry::OpenSMILES::is_chiral_tetrahedral( $atom );
-        next unless @{$atom->{chirality_neighbours}} >= 3 &&
-                    @{$atom->{chirality_neighbours}} <= 4;
+        next unless is_chiral $atom;
 
+        my $has_lone_pair;
         my @chirality_neighbours = @{$atom->{chirality_neighbours}};
-        if( @chirality_neighbours == 3 ) {
+
+        if( Chemistry::OpenSMILES::is_chiral_tetrahedral( $atom ) ||
+            Chemistry::OpenSMILES::is_chiral_planar( $atom ) ) {
+            next unless @chirality_neighbours >= 3 &&
+                        @chirality_neighbours <= 4;
+            $has_lone_pair = @chirality_neighbours == 3;
+        } elsif( Chemistry::OpenSMILES::is_chiral_trigonal_bipyramidal( $atom ) ) {
+            next unless @chirality_neighbours >= 4 &&
+                        @chirality_neighbours <= 5;
+            $has_lone_pair = @chirality_neighbours == 4;
+        }
+
+        if( $has_lone_pair ) {
             @chirality_neighbours = ( $chirality_neighbours[0],
                                       {}, # marking the lone pair
-                                      @chirality_neighbours[1..2] );
-        }
-        if( $atom->{chirality} eq '@' ) {
-            # Reverse the order if counter-clockwise
-            @chirality_neighbours = ( $chirality_neighbours[0],
-                                      reverse @chirality_neighbours[1..3] );
+                                      @chirality_neighbours[1..$#chirality_neighbours] );
         }
 
-        for my $i (0..3) {
-            my $neighbour = $chirality_neighbours[$i];
-            my @chirality_neighbours_now = @chirality_neighbours;
-            
-            if( $i % 2 ) {
-                # Reverse the order due to projected atom change
-                @chirality_neighbours_now = ( $chirality_neighbours_now[0],
-                                              reverse @chirality_neighbours_now[1..3] );
+        if( Chemistry::OpenSMILES::is_chiral_tetrahedral( $atom ) ) {
+            # Algorithm is described in detail in doi:10.1186/s13321-023-00692-1
+            if( $atom->{chirality} eq '@' ) {
+                # Reverse the order if counter-clockwise
+                @chirality_neighbours = ( $chirality_neighbours[0],
+                                          reverse @chirality_neighbours[1..3] );
             }
 
-            my @other = grep { $_ != $neighbour } @chirality_neighbours_now;
-            for my $offset (0..2) {
+            for my $i (0..3) {
+                my $neighbour = $chirality_neighbours[$i];
+                my @chirality_neighbours_now = @chirality_neighbours;
+
+                if( $i % 2 ) {
+                    # Reverse the order due to projected atom change
+                    @chirality_neighbours_now = ( $chirality_neighbours_now[0],
+                                                  reverse @chirality_neighbours_now[1..3] );
+                }
+
+                my @other = grep { $_ != $neighbour } @chirality_neighbours_now;
+                for my $offset (0..2) {
+                    my $connector = {};
+                    $moiety->set_edge_attribute( $neighbour, $connector, 'chiral', 'from' );
+                    $moiety->set_edge_attribute( $atom, $connector, 'chiral', 'to' );
+
+                    $moiety->set_edge_attribute( $connector, $other[0], 'chiral', 1 );
+                    $moiety->set_edge_attribute( $connector, $other[1], 'chiral', 2 );
+                    $moiety->set_edge_attribute( $connector, $other[2], 'chiral', 3 );
+
+                    push @other, shift @other;
+                }
+            }
+        } elsif( Chemistry::OpenSMILES::is_chiral_planar( $atom ) ) {
+            # For square planar environments it is enough to retain the enumeration order of atoms.
+            # To do so, "neighbouring neighbours" are connected together and a link to central atom is placed.
+            if(      $atom->{chirality} eq '@SP2' ) { # 4
+                @chirality_neighbours = map { $chirality_neighbours[$_] } ( 0, 2, 1, 3 );
+            } elsif( $atom->{chirality} eq '@SP3' ) { # Z
+                @chirality_neighbours = map { $chirality_neighbours[$_] } ( 0, 1, 3, 2 );
+            }
+
+            for my $i (0..3) {
                 my $connector = {};
-                $moiety->set_edge_attribute( $neighbour, $connector, 'chiral', 'from' );
-                $moiety->set_edge_attribute( $atom, $connector, 'chiral', 'to' );
+                $moiety->set_edge_attribute( $atom, $connector, 'chiral', 'center' );
+                $moiety->set_edge_attribute( $connector, $chirality_neighbours[$i], 'chiral', 'neighbour' );
+                $moiety->set_edge_attribute( $connector, $chirality_neighbours[($i + 1) % 4], 'chiral', 'neighbour' );
+            }
+        } else { # Trigonal bipyrimidal
+            my $number = substr $atom->{chirality}, 3;
+            my $setting = $TB[$number - 1];
 
-                $moiety->set_edge_attribute( $connector, $other[0], 'chiral', 1 );
-                $moiety->set_edge_attribute( $connector, $other[1], 'chiral', 2 );
-                $moiety->set_edge_attribute( $connector, $other[2], 'chiral', 3 );
+            my @axis  = map  { $chirality_neighbours[$_ - 1] } @{$setting->{axis}};
+            my @other = grep { $_ != $axis[0] && $_ != $axis[1] }
+                        map  { $chirality_neighbours[$_] } 0..4;
+            @other = reverse @other if $setting->{order} eq '@@';
 
-                push @other, shift @other;
+            for my $from (@axis) {
+                my $to = first { $_ != $from } @axis;
+                for my $offset (0..2) {
+                    my $connector = {};
+                    $moiety->set_edge_attribute( $from, $connector, 'chiral', 'from' );
+                    $moiety->set_edge_attribute( $atom, $connector, 'chiral', 'center' );
+                    $moiety->set_edge_attribute( $to,   $connector, 'chiral', 'to' );
+
+                    $moiety->set_edge_attribute( $connector, $other[-1], 'chiral', 'counter-clockwise' );
+                    $moiety->set_edge_attribute( $connector, $other[ 1], 'chiral', 'clockwise' );
+
+                    push @other, shift @other;
+                }
+                @other = reverse @other; # Inverting the axis
             }
         }
     }
@@ -298,14 +371,14 @@ sub cis_trans_to_pseudoedges
         next if @atom2_neighbours < 2 || @atom2_neighbours > 3 ||
                 @atom3_neighbours < 2 || @atom3_neighbours > 3;
 
-        my( $atom1 ) = grep { is_cis_trans_bond( $moiety, $atom2, $_ ) }
-                            @atom2_neighbours;
-        my( $atom4 ) = grep { is_cis_trans_bond( $moiety, $atom3, $_ ) }
-                            @atom3_neighbours;
+        my $atom1 = first { is_cis_trans_bond( $moiety, $atom2, $_ ) }
+                          @atom2_neighbours;
+        my $atom4 = first { is_cis_trans_bond( $moiety, $atom3, $_ ) }
+                          @atom3_neighbours;
         next unless $atom1 && $atom4;
 
-        my( $atom1_para ) = grep { $_ != $atom1 && $_ != $atom3 } @atom2_neighbours;
-        my( $atom4_para ) = grep { $_ != $atom4 && $_ != $atom2 } @atom3_neighbours;
+        my $atom1_para = first { $_ != $atom1 && $_ != $atom3 } @atom2_neighbours;
+        my $atom4_para = first { $_ != $atom4 && $_ != $atom2 } @atom3_neighbours;
 
         my $is_cis = $moiety->get_edge_attribute( $atom1, $atom2, 'bond' ) ne
                      $moiety->get_edge_attribute( $atom3, $atom4, 'bond' );
